
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include "initng_readahead_makecache.h"
#include "initng_readahead_procparser.h"
#include "initng_readahead_database.h"

static int scan_proc_maps(int pid)
{
    FILE *fp;
    char buf[256];
    char *path;
    char *newline;

    snprintf(buf, sizeof(buf), PROCDIR "/%d/maps", pid);
    if ((fp = fopen(buf, "r")) == NULL)
    {
        /* Not critical, we're forking like crazy during boot, this WILL happen F_("Failed to open %s\n",buf); */
        return (FALSE);
    }
    while (fgets(buf, sizeof(buf), fp))
    {
        /* 2aaaab123000-2aaaab1eb000 r-xp 00000000 fe:02 317498                     /usr/lib64/libasound.so.2.0.0 */
        /* Turn that into the path part */
        path = rindex(buf, ' ');
        if (path)
            path++;
        newline = index(path, '\n');
        if (newline)
            newline[0] = 0;
        /* Line formated */
        add_to_db(path);
    }
    fclose(fp);
    return (TRUE);
}

static int scan_proc_task_fds(int pid, char *thread)
{
    DIR *dd;
    struct dirent *d;
    int readlink_ret;
    char buf[256];
    char readlink_buf[MAXPATH];

    snprintf(buf, sizeof(buf), PROCDIR "/%d/task/%s/fd", pid, thread);
    if ((dd = opendir(buf)) == NULL)
    {
        /* Not critical, we're forking like crazy during boot, this WILL happen F_("Failed to open %s\n",buf); */
        return (FALSE);
    }
    while ((d = readdir(dd)) != NULL)
    {
        if (d->d_name && d->d_name[0] != '.')
        {
            snprintf(buf, sizeof(buf), PROCDIR "/%d/task/%s/fd/%s", pid,
                     thread, d->d_name);
            readlink_ret = readlink(buf, readlink_buf, MAXPATH);
            if (readlink_ret > 0 && readlink_ret < MAXPATH)
            {
                readlink_buf[readlink_ret] = 0;
                add_to_db(readlink_buf);    /* todo, use add_to_db_real and pass readlink_ret as length? TINY performance increase... */
                D_("Readlink : %s\n", readlink_buf);
            }
            else
            {
                F_("Readlink on pid %d, thread %s, name %s failed (returned %d)\n", pid, thread, d->d_name, readlink_ret);
                if (readlink_ret == -1)
                {
                    break;
                    /*F_("Strerror : %s\n",strerror(errno)); */
                }
            }
        }
    }
    closedir(dd);
    return (TRUE);
}

static int scan_proc_task_pids(int pid)
{
    DIR *dd;
    struct dirent *d;
    char buf[256];

    snprintf(buf, sizeof(buf), PROCDIR "/%d/task", pid);
    if ((dd = opendir(buf)) == NULL)
    {
        /* Not critical, we're forking like crazy during boot, this WILL happen F_("Failed to open %s\n",buf); */
        return (FALSE);
    }
    while ((d = readdir(dd)) != NULL)
    {
        if (d->d_name && d->d_name[0] != '.')
            if (!scan_proc_task_fds(pid, d->d_name))
                return (FALSE);
    }
    closedir(dd);
    return (TRUE);
}

static int get_proc_cmdline(int pid)
{
    char buf[256];
    FILE *fp;
    char *firstarg;

    snprintf(buf, sizeof(buf), PROCDIR "/%d/cmdline", pid);
    if ((fp = fopen(buf, "r")) == NULL)
    {
        /* Not critical, we're forking like crazy during boot, this WILL happen F_("Failed to open %s\n",buf); */
        return (FALSE);
    }
    if (fgets(buf, sizeof(buf), fp))
    {
        firstarg = index(buf, ' ');
        if (firstarg)
            firstarg[0] = 0;
        add_to_db(buf);
    }
    fclose(fp);
    return (TRUE);
}

int scan_proc(void)
{
    DIR *dd;
    struct dirent *d;
    int pid;

    if ((dd = opendir(PROCDIR)) == NULL)
    {
        F_("Failed to open %s\n", PROCDIR);
        return FALSE;
    }
    while ((d = readdir(dd)) != NULL)
    {
        pid = atoi(d->d_name);
        if (pid > 1)
        {
            D_("Pid number : %d\n", pid);
            /* First scan /proc/PID/maps */
            if (!scan_proc_maps(pid))
                continue;
            // Then scan /proc/PID/task/*/fd 
            if (!scan_proc_task_pids(pid))
                continue;
            /* Then grab /proc/PID/cmdline[0] */
            get_proc_cmdline(pid);
        }
    }
    closedir(dd);
    return (TRUE);
}
